#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "dizzy.h"

int decompress(char *filename, char *filenameout, char *filenameCompr, u32 address, u32 count)
{
	FILE  *fd, *fdout, *fdoutcompr = NULL;
	u8    tile16[32], tile16dec[32];
	u32   adr_cnt;
	u32   cmprline_len, unc_bytes, run_bytes;
	if ((fd = fopen(filename, "rb")) != NULL)
	{
		if ((fdout = fopen(filenameout, "wb")) == NULL)
		{
			printf("Oj bida, bida... nie dalo rady pliku otworzyc \"%s\"!\n", filenameout);
			fclose(fd);
			return 0;
		}
		if (filenameCompr != NULL)
		{
			if ((fdoutcompr = fopen(filenameCompr, "wb")) == NULL)
			{
				printf("Oj bida, bida... nie dalo rady pliku otworzyc \"%s\"!\n", filenameCompr);
				fclose(fd);
				fclose(fdout);
				return 0;
			}
		}
		adr_cnt = address;
		fseek(fd, address, SEEK_SET);
		for (u32 i = 0; i < count; i++)
		{
			memset(tile16, 0, 32);
			memset(tile16dec, 0, 32);
			if (!fread(tile16, 1, 1, fd))
			{
				printf("Blad na linii nr: %d, adres: 0x%08x\n", i+1, adr_cnt);
				break;
			}
			unc_bytes = (((tile16[0]>>4) & 0x0F) + 1);
			run_bytes = (tile16[0] & 0x0F);
			if (((run_bytes + unc_bytes) > 16))
			{
				printf("Mamy problem: dlugosc linii wieksza od 16!\nNumer linii: %d, adres: 0x%08x\n", i+1, adr_cnt);
				break;
			}
			printf("tile16[0] = 0x%02x\n", tile16[0]);
			cmprline_len = 17 - run_bytes;
			printf("Line Len: %d, address: 0x%08x\n", cmprline_len, adr_cnt);
			fread(&tile16[1], cmprline_len-1, 1, fd);
			adr_cnt += cmprline_len;

			memcpy(tile16dec, &tile16[1], unc_bytes);
			memset(&tile16dec[unc_bytes], tile16[unc_bytes], run_bytes);
			memcpy(&tile16dec[unc_bytes+run_bytes], &tile16[unc_bytes+1], 16-unc_bytes-run_bytes);

			fwrite(tile16dec, 16, 1, fdout);
			if (fdoutcompr) fwrite(tile16, cmprline_len, 1, fdoutcompr);
		}
		if (fdoutcompr) fclose(fdoutcompr);
		fclose(fd);
		fclose(fdout);
	} else
	{
		printf("Oj bida, bida... nie dalo rady pliku otworzyc \"%s\"!\n", filename);
	}
	return 1;
}
int compress(char *filenameIn, char *filenameOut)
{
	FILE *fdin, *fdout;
	char tile16dec[16], tile16[32];
	u32 maxRun, currentRun, maxPos, currentPos, runPending;
	if ((fdin = fopen(filenameIn, "rb")) == NULL)
	{
		printf("Oj bida, bida... nie dalo rady pliku otworzyc \"%s\"!\n", filenameIn);
		return 0;
	}
	if ((fdout = fopen(filenameOut, "wb")) == NULL)
	{
		fclose(fdin);
		printf("Oj bida, bida... nie dalo rady pliku otworzyc \"%s\"!\n", filenameOut);
		return 0;
	}
	memset(tile16, 0, 32);
	memset(tile16dec, 0, 16);
	while (fread(tile16dec, 16, 1, fdin))
	{
		maxRun = currentRun = maxPos = currentPos = 0;
		runPending = 0;
		for (u32 i = 1; i < 16; i++)
		{
			if (tile16dec[i-1] == tile16dec[i])
			{
				if (runPending == 0) currentPos = i-1;
				currentRun++;
				runPending = 1;
			} else
			{
				runPending = 0;
				if (currentRun >= maxRun)
				{
					maxPos = currentPos;
					maxRun = currentRun;
				}
				currentPos = currentRun = 0;
			}
		}
		if (currentRun >= maxRun)
		{
			maxPos = currentPos;
			maxRun = currentRun;
		}

		if ((maxPos == maxRun) && (maxRun == 0))
		{
			maxPos = 15;
		}
		tile16[0] = ((maxPos << 4) & 0xF0) | (maxRun & 0x0F);
		memcpy(&tile16[1], tile16dec, maxPos);
		memcpy(&tile16[maxPos+1], &tile16dec[maxPos+maxRun], 17-maxPos-maxRun);
		fwrite(tile16, 17-maxRun, 1, fdout);
		memset(tile16, 0, 32);
		memset(tile16dec, 0, 16);
	}
	fclose(fdin);
	fclose(fdout);
	return 1;
}
void printUsage()
{
	printf(
		"dizzy pack/unpack v0.1, by ffgriever\n\n"
		"dizzy.exe [opcje]\n\n"
		"Opcje:\n"
		"   -c              Kompresuj wskazany plik\n\n"
		"   -d              Dekompresuj wskazany plik\n\n"
		"   -in=plik        Sciezka do pliku wejsciowego\n\n"
		"   -out=plik       Sciezka do pliku wyjsciowego\n\n"
		"   -compr=plik     Sciezka do pliku skompresowanego\n"
		"                   zapisywanego w trakcie dekompresji\n"
		"                   parametr jest opcjonalny\n\n"
		"   -addr=liczba    adres od ktorego zaczac dekompresje\n\n"
		"   -cnt=liczba     liczba linii do wypakowania\n\n"
		);
}
int main(int argc, char **argv)
{
	//decompress("dizzy.nes", "unc_chr.chr", "comp_chr.chr", 0x00027A12, 64);
	//compress("unc_chr.chr", "rec_chr.chr");
	//decompress("comp_chr.chr", "3333unc_chr.chr", NULL, 0, 64);
	u32 operation = 0;
	char *inFile = NULL;
	char *outFile = NULL;
	char *outFileCompr = NULL;
	u32 address = 0xFFFFFFFF;
	u32 count = 0xFFFFFFFF;
	int ret = 0;
	for (int i = 1; i < argc; i++)
	{
		if (argv[i][0] == '-')
		{
			if (strncmp(argv[i], "-in=", 4) == 0)
			{
				inFile = &argv[i][4];
			} else if (strncmp(argv[i], "-out=", 5) == 0)
			{
				outFile = &argv[i][5];
			} else if (strncmp(argv[i], "-compr=", 7) == 0)
			{
				outFileCompr = &argv[i][7];
			} else if (strncmp(argv[i], "-addr=", 6) == 0)
			{
				address = strtol(&argv[i][6],NULL, 0);
			} else if (strncmp(argv[i], "-cnt=", 5) == 0)
			{
				count = strtol(&argv[i][5],NULL, 0);
			} else if (strncmp(argv[i], "-c", 2) == 0)
			{
				if (operation > 0)
				{
					printf("Operacja do wykonania zostala juz podana!\n\n");
					printUsage();
				}
				operation = OPER_COMPRESS;
			} else if (strncmp(argv[i], "-d", 2) == 0)
			{
				if (operation > 0)
				{
					printf("Operacja do wykonania zostala juz podana!\n\n");
					printUsage();
				}
				operation = OPER_DECOMPRESS;
			}
		}
	}
	if (inFile == NULL || outFile == NULL || operation == 0)
	{
		printf("Za malo parametrow!\n\n");
		printUsage();
		return 0;
	}
	switch (operation)
	{
		case OPER_COMPRESS:
			ret = compress(inFile, outFile);
			break;
		case OPER_DECOMPRESS:
			if (address == 0xFFFFFFFF || count == 0xFFFFFFFF)
			{
				printf("Za malo parametrow dla wskazanej operacji - konieczny adres i liczba!\n\n");
				printUsage();
				return 0;
			}
			ret = decompress(inFile, outFile, outFileCompr, address, count);
			break;
	}
	if (ret)
	{
		printf("\nOperacja zakonczona sukcesem!\n");
	} else
	{
		printf("\nOperacja zakonczona niepowodzeniem!\n");
	}
	return ret;
}
